@@ -38,16 +38,18 @@
 #include "udscs.h"
 #include "vdagentd-proto.h"
 #include "audio.h"
-#include "x11.h"
 #include "file-xfers.h"
 #include "clipboard.h"
+#include "display.h"
+
+#define MAX_RETRY_CONNECT_SYSTEM_AGENT 60
 
 typedef struct VDAgent {
     VDAgentClipboards *clipboards;
-    struct vdagent_x11 *x11;
+    VDAgentDisplay *display;
     struct vdagent_file_xfers *xfers;
-    struct udscs_connection *conn;
-    GIOChannel *x11_channel;
+    UdscsConnection *conn;
+    gint udscs_num_retry;
 
     GMainLoop *loop;
 } VDAgent;
@@ -66,25 +68,32 @@ static gchar *portdev = NULL;
 static gchar *vdagentd_socket = NULL;
 
 static GOptionEntry entries[] = {
-    { "debug", 'd', 0,
-       G_OPTION_ARG_NONE, &debug,
-       "Enable debug", NULL },
-    { "virtio-serial-port-path", 's', 0,
+    { "debug", 'd',
+      G_OPTION_FLAG_NONE,
+      G_OPTION_ARG_NONE, &debug,
+      "Enable debug", NULL },
+    { "virtio-serial-port-path", 's',
+      G_OPTION_FLAG_NONE,
       G_OPTION_ARG_STRING, &portdev,
       "Set virtio-serial path ("  DEFAULT_VIRTIO_PORT_PATH ")", NULL },
-    { "vdagentd-socket", 'S', 0, G_OPTION_ARG_STRING,
-       &vdagentd_socket,
-       "Set spice-vdagentd socket (" VDAGENTD_SOCKET ")", NULL },
-    { "foreground", 'x', G_OPTION_FLAG_REVERSE,
-       G_OPTION_ARG_NONE, &do_daemonize,
-       "Do not daemonize the agent", NULL },
-    { "file-xfer-save-dir", 'f', 0,
+    { "vdagentd-socket", 'S',
+      G_OPTION_FLAG_NONE,
+      G_OPTION_ARG_STRING, &vdagentd_socket,
+      "Set spice-vdagentd socket (" VDAGENTD_SOCKET ")", NULL },
+    { "foreground", 'x',
+      G_OPTION_FLAG_REVERSE,
+      G_OPTION_ARG_NONE, &do_daemonize,
+      "Do not daemonize the agent", NULL },
+    { "file-xfer-save-dir", 'f',
+      G_OPTION_FLAG_NONE,
       G_OPTION_ARG_STRING, &fx_dir,
       "Set directory to file transfers files", "<dir|xdg-desktop|xdg-download>"},
-    { "file-xfer-open-dir", 'o', 0,
-       G_OPTION_ARG_INT, &fx_open_dir,
-       "Open directory after completing file transfer", "<0|1>" },
-    { "x11-abort-on-error", 'y', G_OPTION_FLAG_HIDDEN,
+    { "file-xfer-open-dir", 'o',
+      G_OPTION_FLAG_NONE,
+      G_OPTION_ARG_INT, &fx_open_dir,
+      "Open directory after completing file transfer", "<0|1>" },
+    { "x11-abort-on-error", 'y',
+      G_OPTION_FLAG_HIDDEN,
       G_OPTION_ARG_NONE, &x11_sync,
       "Aborts on errors from X11", NULL },
     { NULL }
@@ -107,7 +116,7 @@ static const gchar *xfer_get_download_directory(VDAgent *agent)
         return fx_dir;
     }
 
-    return g_get_user_special_dir(vdagent_x11_has_icons_on_desktop(agent->x11) ?
+    return g_get_user_special_dir(vdagent_display_has_icons_on_desktop(agent->display) ?
                                   G_USER_DIRECTORY_DESKTOP :
                                   G_USER_DIRECTORY_DOWNLOAD);
 }
@@ -137,7 +146,7 @@ static gboolean vdagent_init_file_xfer(VDAgent *agent)
     }
 
     open_dir = fx_open_dir == -1 ?
-               !vdagent_x11_has_icons_on_desktop(agent->x11) :
+               !vdagent_display_has_icons_on_desktop(agent->display) :
                fx_open_dir;
 
     agent->xfers = vdagent_file_xfers_create(agent->conn, xfer_dir,
@@ -158,21 +167,21 @@ static void vdagent_quit_loop(VDAgent *agent)
 {
     /* other GMainLoop(s) might be running, quit them before agent->loop */
     if (agent->clipboards) {
-        vdagent_clipboards_finalize(agent->clipboards, agent->conn != NULL);
-        agent->clipboards = NULL;
+        vdagent_clipboards_set_conn(agent->clipboards, agent->conn);
+        g_clear_object(&agent->clipboards);
     }
     if (agent->loop)
         g_main_loop_quit(agent->loop);
 }
 
-static void daemon_read_complete(struct udscs_connection **connp,
+static void daemon_read_complete(UdscsConnection *conn,
     struct udscs_message_header *header, uint8_t *data)
 {
-    VDAgent *agent = udscs_get_user_data(*connp);
+    VDAgent *agent = g_object_get_data(G_OBJECT(conn), "agent");
 
     switch (header->type) {
     case VDAGENTD_MONITORS_CONFIG:
-        vdagent_x11_set_monitor_config(agent->x11, (VDAgentMonitorsConfig *)data, 0);
+        vdagent_display_set_monitor_config(agent->display, (VDAgentMonitorsConfig *)data, 0);
         break;
     case VDAGENTD_CLIPBOARD_REQUEST:
         vdagent_clipboard_request(agent->clipboards, header->arg1, header->arg2);
@@ -201,7 +210,7 @@ static void daemon_read_complete(struct udscs_connection **connp,
             vdagent_file_xfers_start(agent->xfers,
                                      (VDAgentFileXferStartMessage *)data);
         } else {
-            vdagent_file_xfers_error_disabled(*connp,
+            vdagent_file_xfers_error_disabled(conn,
                                               ((VDAgentFileXferStartMessage *)data)->id);
         }
         break;
@@ -210,7 +219,7 @@ static void daemon_read_complete(struct udscs_connection **connp,
             vdagent_file_xfers_status(agent->xfers,
                                       (VDAgentFileXferStatusMessage *)data);
         } else {
-            vdagent_file_xfers_error_disabled(*connp,
+            vdagent_file_xfers_error_disabled(conn,
                                               ((VDAgentFileXferStatusMessage *)data)->id);
         }
         break;
@@ -222,11 +231,14 @@ static void daemon_read_complete(struct udscs_connection **connp,
         break;
     case VDAGENTD_AUDIO_VOLUME_SYNC: {
         VDAgentAudioVolumeSync *avs = (VDAgentAudioVolumeSync *)data;
+        uint16_t *volume = g_memdup2(avs->volume, sizeof(uint16_t) * avs->nchannels);
+
         if (avs->is_playback) {
-            vdagent_audio_playback_sync(avs->mute, avs->nchannels, avs->volume);
+            vdagent_audio_playback_sync(avs->mute, avs->nchannels, volume);
         } else {
-            vdagent_audio_record_sync(avs->mute, avs->nchannels, avs->volume);
+            vdagent_audio_record_sync(avs->mute, avs->nchannels, volume);
         }
+        g_free(volume);
         break;
     }
     case VDAGENTD_FILE_XFER_DATA:
@@ -234,12 +246,12 @@ static void daemon_read_complete(struct udscs_connection **connp,
             vdagent_file_xfers_data(agent->xfers,
                                     (VDAgentFileXferDataMessage *)data);
         } else {
-            vdagent_file_xfers_error_disabled(*connp,
+            vdagent_file_xfers_error_disabled(conn,
                                               ((VDAgentFileXferDataMessage *)data)->id);
         }
         break;
     case VDAGENTD_GRAPHICS_DEVICE_INFO:
-        vdagent_x11_handle_graphics_device_info(agent->x11, data, header->size);
+        vdagent_display_handle_graphics_device_info(agent->display, data, header->size);
         break;
     case VDAGENTD_CLIENT_DISCONNECTED:
         vdagent_clipboards_release_all(agent->clipboards);
@@ -253,10 +265,15 @@ static void daemon_read_complete(struct udscs_connection **connp,
     }
 }
 
-static void daemon_disconnect_cb(struct udscs_connection *conn)
+static void daemon_error_cb(VDAgentConnection *conn, GError *err)
 {
-    VDAgent *agent = udscs_get_user_data(conn);
-    agent->conn = NULL;
+    VDAgent *agent = g_object_get_data(G_OBJECT(conn), "agent");
+
+    if (err) {
+        syslog(LOG_ERR, "%s", err->message);
+        g_error_free(err);
+    }
+    g_clear_pointer(&agent->conn, vdagent_connection_destroy);
     vdagent_quit_loop(agent);
 }
 
@@ -320,15 +337,7 @@ static int daemonize(void)
     return 0;
 }
 
-static gboolean x11_io_channel_cb(GIOChannel *source,
-                                  GIOCondition condition,
-                                  gpointer data)
-{
-    VDAgent *agent = data;
-    vdagent_x11_do_read(agent->x11);
 
-    return G_SOURCE_CONTINUE;
-}
 
 gboolean vdagent_signal_handler(gpointer user_data)
 {
@@ -354,13 +363,12 @@ static VDAgent *vdagent_new(void)
 static void vdagent_destroy(VDAgent *agent)
 {
     vdagent_finalize_file_xfer(agent);
-    vdagent_x11_destroy(agent->x11, agent->conn == NULL);
-    udscs_destroy_connection(&agent->conn);
+    vdagent_display_destroy(agent->display, agent->conn == NULL);
+    g_clear_pointer(&agent->conn, vdagent_connection_destroy);
 
     while (g_source_remove_by_user_data(agent))
         continue;
 
-    g_clear_pointer(&agent->x11_channel, g_io_channel_unref);
     g_clear_pointer(&agent->loop, g_main_loop_unref);
     g_free(agent);
 }
@@ -368,32 +376,50 @@ static void vdagent_destroy(VDAgent *agent)
 static gboolean vdagent_init_async_cb(gpointer user_data)
 {
     VDAgent *agent = user_data;
+    GError *err = NULL;
 
     agent->conn = udscs_connect(vdagentd_socket,
-                                daemon_read_complete, daemon_disconnect_cb,
-                                debug);
+                                daemon_read_complete,
+                                daemon_error_cb,
+                                debug,
+                                &err);
     if (agent->conn == NULL) {
+        if (agent->udscs_num_retry == MAX_RETRY_CONNECT_SYSTEM_AGENT) {
+            syslog(LOG_WARNING,
+                   "Failed to connect to spice-vdagentd at %s (tried %d times)",
+                   vdagentd_socket, agent->udscs_num_retry);
+            g_error_free(err);
+            goto err_init;
+        }
+        if (agent->udscs_num_retry == 0) {
+            /* Log only when it fails and at the end */
+            syslog(LOG_DEBUG,
+                   "Failed to connect with spice-vdagentd due '%s'. Trying again in 1s",
+                   err->message);
+        }
+        g_error_free(err);
+        agent->udscs_num_retry++;
         g_timeout_add_seconds(1, vdagent_init_async_cb, agent);
         return G_SOURCE_REMOVE;
     }
-    udscs_set_user_data(agent->conn, agent);
+    if (agent->udscs_num_retry != 0) {
+        syslog(LOG_DEBUG,
+               "Connected with spice-vdagentd after %d attempts",
+               agent->udscs_num_retry);
+    }
+    agent->udscs_num_retry = 0;
+    g_object_set_data(G_OBJECT(agent->conn), "agent", agent);
 
-    agent->x11 = vdagent_x11_create(agent->conn, debug, x11_sync);
-    if (agent->x11 == NULL)
-        goto err_init;
-    agent->x11_channel = g_io_channel_unix_new(vdagent_x11_get_fd(agent->x11));
-    if (agent->x11_channel == NULL)
+    agent->display = vdagent_display_create(agent->conn, debug, x11_sync);
+    if (agent->display == NULL)
         goto err_init;
 
-    g_io_add_watch(agent->x11_channel,
-                   G_IO_IN,
-                   x11_io_channel_cb,
-                   agent);
 
     if (!vdagent_init_file_xfer(agent))
         syslog(LOG_WARNING, "File transfer is disabled");
 
-    agent->clipboards = vdagent_clipboards_init(agent->x11, agent->conn);
+    agent->clipboards = vdagent_clipboards_new(vdagent_display_get_x11(agent->display));
+    vdagent_clipboards_set_conn(agent->clipboards, agent->conn);
 
     if (parent_socket != -1) {
         if (write(parent_socket, "OK", 2) != 2)
@@ -415,7 +441,8 @@ int main(int argc, char *argv[])
     GOptionContext *context;
     GError *error = NULL;
     VDAgent *agent;
-    char **orig_argv = g_memdup(argv, sizeof(char*) * (argc+1));
+    char **orig_argv = g_memdup2(argv, sizeof(char*) * (argc+1));
+    orig_argv[argc] = NULL; /* To avoid clang analyzer false-positive */
 
     context = g_option_context_new(NULL);
     g_option_context_add_main_entries(context, entries, NULL);
@@ -423,7 +450,9 @@ int main(int argc, char *argv[])
                                  "\tSpice session guest agent: X11\n"
                                  "\tVersion: " VERSION);
 #ifdef WITH_GTK
+#if ! GTK_CHECK_VERSION(3, 98, 0)  // not with GTK4 - option parsing is gone
     g_option_context_add_group(context, gtk_get_option_group(FALSE));
+#endif
 #endif
     g_option_context_parse(context, &argc, &argv, &error);
     g_option_context_free(context);
@@ -431,6 +460,7 @@ int main(int argc, char *argv[])
     if (error != NULL) {
         g_printerr("Invalid arguments, %s\n", error->message);
         g_clear_error(&error);
+        g_free(orig_argv);
         return -1;
     }
 
@@ -445,7 +475,8 @@ int main(int argc, char *argv[])
             LOG_USER);
 
     if (!g_file_test(portdev, G_FILE_TEST_EXISTS)) {
-        g_debug("vdagent virtio channel %s does not exist, exiting", portdev);
+        syslog(LOG_DEBUG, "vdagent virtio channel %s does not exist, exiting", portdev);
+        g_free(orig_argv);
         return 0;
     }
 
@@ -455,9 +486,16 @@ int main(int argc, char *argv[])
     syslog(LOG_INFO, "vdagent started");
 
 #ifdef WITH_GTK
+#if GTK_CHECK_VERSION(3, 98, 0)
+    // GTK4: use Wayland if possible, otherwise use X11
+    gdk_set_allowed_backends("wayland,x11");
+    gtk_init();
+#else
+    // GTK3: need to use X11 as talking to Wayland will cause bugs
     gdk_set_allowed_backends("x11");
     gtk_init(NULL, NULL);
 #endif
+#endif
 
 reconnect:
     if (version_mismatch) {
@@ -475,6 +513,9 @@ reconnect:
     vdagent_destroy(agent);
     agent = NULL;
 
+    /* allow the VDAgentConnection to finalize properly */
+    g_main_context_iteration(NULL, FALSE);
+
     if (!quit && do_daemonize)
         goto reconnect;
 
