@@ -28,7 +28,7 @@
 
 struct vdagent_x11;
 
-struct vdagent_x11 *vdagent_x11_create(struct udscs_connection *vdagentd,
+struct vdagent_x11 *vdagent_x11_create(UdscsConnection *vdagentd,
     int debug, int sync);
 void vdagent_x11_destroy(struct vdagent_x11 *x11, int vdagentd_disconnected);
 
@@ -38,7 +38,7 @@ void vdagent_x11_do_read(struct vdagent_x11 *x11);
 void vdagent_x11_set_monitor_config(struct vdagent_x11 *x11,
     VDAgentMonitorsConfig *mon_config, int fallback);
 
-#ifndef WITH_GTK
+#ifndef USE_GTK_FOR_CLIPBOARD
 void vdagent_x11_clipboard_grab(struct vdagent_x11 *x11, uint8_t selection,
     uint32_t *types, uint32_t type_count);
 void vdagent_x11_clipboard_request(struct vdagent_x11 *x11,
@@ -50,7 +50,10 @@ void vdagent_x11_clipboard_release(struct vdagent_x11 *x11, uint8_t selection);
 void vdagent_x11_client_disconnected(struct vdagent_x11 *x11);
 #endif
 
-int vdagent_x11_has_icons_on_desktop(struct vdagent_x11 *x11);
-void vdagent_x11_handle_graphics_device_info(struct vdagent_x11 *x11, uint8_t *data, size_t size);
+gchar *vdagent_x11_get_wm_name(struct vdagent_x11 *x11);
+
+void vdagent_x11_handle_device_display_info(struct vdagent_x11 *x11,
+                                            VDAgentDeviceDisplayInfo *device_display_info,
+                                            gboolean has_virtual_zero_display);
 
 #endif
